/*
 * hb_data_log.c
 *
 * Created: 04.07.2019 17:02:16
 *  Author: user
 */

#include "hb_data_log.h"
#include "Utils/hb_utils.h"       // crc16()
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
//#include <malloc.h>

#include "hb_hal.h"

    #define max(a,b) \
        ({ __typeof__ (a) _a = (a); \
            __typeof__ (b) _b = (b); \
            _a > _b ? _a : _b; })

    #define min(a,b) \
        ({ __typeof__ (a) _a1 = (a); \
            __typeof__ (b) _b1 = (b); \
            _a1 < _b1 ? _a1 : _b1; })


/*************************************************************************
*  Implementation of the HB_DATA_LOG related methods
*************************************************************************/

// AK 2016-10-08: version 1.0
#define DATA_LOG_VERSION_MAJOR  1
#define DATA_LOG_VERSION_MINOR  0

#if (MAX_DATA_FILE_ID > TOTAL_DATA_LOGS)
#error "MAX_DATA_FILE_ID can't be greater than TOTAL_DATA_LOGS"
#endif

#if (TOTAL_DATA_LOGS != 32)
#error "Are you sure? Do you understand all the consequences?"
#endif

// NOTE: in order to minimize file read/write operation we always save the whole section
//       of a global .cfg-file that contains an array of number of records available
//       in individual .dat-files. In addition, we associate with each .dat file a section,
//       in the global .cfg-file that may store some .dat-file specific meta-information.
//       In order to achieve the both mentioned goals we use the following structure for
//       the global .cfg-file:

#define PER_FILE_METADATA_BYTES 128     // AK TODO: should we use 256 bytes?

#define DATA_LOG_HEADER_BYTES   128


typedef struct HB_PACKED data_log_metadata_
{
    uint8_t data_buf[PER_FILE_METADATA_BYTES - 2];
    uint16_t crc16;

} DATA_LOG_METADATA;

STATIC_ASSERT_TYPE_SIZE(DATA_LOG_METADATA, PER_FILE_METADATA_BYTES);


/* DATA_LOG_INFO describes the format of the global .cfg-file and is stored in the metadata[0] section */
typedef struct data_log_format_info_
{
    uint8_t  version_major;
    uint8_t  version_minor;
    uint8_t  max_data_logs;
    uint8_t  size;                // of DATA_LOG_INFO
    uint16_t size_of_metadata;    // size of module PER_FILE_METADATA_BYTES
    uint16_t number_of_tables;    // number of (sections with) per-file data tables

    /* --- 8-bytes boundary --------------- */

    uint8_t  reserved2[DATA_LOG_HEADER_BYTES - 8 - 2];
    uint16_t crc16;

} DATA_LOG_HEADER;

STATIC_ASSERT_TYPE_SIZE(DATA_LOG_HEADER, DATA_LOG_HEADER_BYTES);


typedef struct HB_PACKED data_log_cfg_file_
{
    DATA_LOG_HEADER header;

    /* --- a new section of flash memory ----- */

    uint32_t file_offsets[TOTAL_DATA_LOGS];
    uint16_t file_offset_crc16;
    uint16_t pad1;

    /* --- a new section of flash memory ----- */

    uint32_t file_sizes[TOTAL_DATA_LOGS];
    uint16_t file_size_crc16;
    uint16_t pad2;

    /* --- a new section of flash memory ----- */

    uint32_t recs_counts[TOTAL_DATA_LOGS];
    uint16_t recs_count_crc16;
    uint16_t pad3;

    /* --- a new section of flash memory ----- */

    uint32_t min_valid_rec_idx[TOTAL_DATA_LOGS];
    uint16_t min_valid_rec_idx_crc16;
    uint16_t pad4;

    /* --- a new section of flash memory ----- */

    uint32_t reserved_table_2[TOTAL_DATA_LOGS];
    uint16_t reserved_table_2_crc16;
    uint16_t pad5;

    /* --- a new section of flash memory ----- */

    uint32_t reserved_table_3[TOTAL_DATA_LOGS];
    uint16_t reserved_table_3_crc16;
    uint16_t pad6;

    /* --- a new section of flash memory per metadata item ----- */
    DATA_LOG_METADATA metadata[TOTAL_DATA_LOGS];

} DATA_LOG_CFG_FILE;


#define NUMBER_OF_PER_FILE_TABLES     6

#define CFG_FILE_SIZE           (sizeof(DATA_LOG_CFG_FILE))

// AK TODO: for the raw access to the flash memory each logical section of the .cfg-file
//          may need to be placed in a separate sector (or even block) of a flash memory
//          so that we could easily implement write counters and hot replacement of the
//          worn sectors

// #define FLASH_SECTOR_SIZE_BYTES        256
// #define OFFSET_OF_RECS_SIZE_SECTION    0
// #define OFFSET_OF_METADATA_SECTION     FLASH_SECTOR_SIZE_BYTES
//
// #define CFG_FILE_SIZE        (FLASH_SECTOR_SIZE_BYTES * (TOTAL_DATA_LOGS + 1))

#define DEFAULT_RECS_COUNTS_OFFSET          offsetof(DATA_LOG_CFG_FILE, recs_counts)
#define DEFAULT_MIN_VALID_REC_IDX_OFFSET    offsetof(DATA_LOG_CFG_FILE, min_valid_rec_idx)
#define DEFAULT_METADATA_OFFSET             offsetof(DATA_LOG_CFG_FILE, metadata)
#define SECTION_SIZE                        sizeof(DATA_LOG_METADATA)


#ifdef HB_ATMEL_ARM
#define FNAME_PREFIX  "0:"   // platform-dependent prefix (drive & path)
#elif defined(HB_WIN32)
#define FNAME_PREFIX  ""
#else
#error "Please define FNAME_PREFIX for your platform"
#endif

#define DAT_FILE_EXT  ".dat"
#define CFG_FILE_EXT  ".cfg"

#define MAX_FNAME_LEN 16     // FNAME_PREFIX + bare file name + FILE_EXT + '/0'


#define DATA_LOGS_CFG_FILE_NAME    "datalogs"
#define DATA_FILE_ID_DATA_LOG_CFG  0


// a handle of the data log global .cfg-file for storing micro-FS information:
// the number of records in the .dat files and other per-file info
static f_handle_t _cfg_file = {0};
static uint32_t _recs_counts_cache[MAX_DATA_FILE_ID];
static uint32_t _min_valid_rec_idx_cache[MAX_DATA_FILE_ID];


/* just combines the provided zero-ended strings. The caller is responsible for handling any buffer overflow issues */
static void _compose_fname(const char* prefix, const char* bare_fname, const char* suffix, char* buf)
{
    size_t offset = 0;
    size_t len;

    len = strlen(prefix);
    hal_memcpy(&buf[offset], prefix, len);
    offset += len;

    len = strlen(bare_fname);
    hal_memcpy(&buf[offset], bare_fname, len);
    offset += len;

    len = strlen(suffix);
    hal_memcpy(&buf[offset], suffix, len);
    offset += len;

    buf[offset] = 0;
}
/* --------------------------------------------------------------- */


// AK TODO: do we need to implement this kind of protection or rely on a caller instead?
// static uint32_t _data_log_is_inited(void)
// {
//     return (_cfg_file.handle != 0) ? HB_TRUE : HB_FALSE;
// }
/* --------------------------------------------------------------- */


static uint32_t _read_rec_counts_table(void)
{
    uint32_t bytes_to_read = sizeof(_recs_counts_cache);
    uint32_t bytes_read;

    hb_stor_set_rw_pos(&_cfg_file, DEFAULT_RECS_COUNTS_OFFSET);
    bytes_read = hb_stor_read_buf(&_cfg_file, &_recs_counts_cache[0], bytes_to_read);

    // AK TODO: check crc16 for this section? How will it affect performance?

    return (bytes_read == bytes_to_read) ? HB_SUCCESS : HB_FAIL;
}
/* --------------------------------------------------------------- */


static uint32_t _write_rec_counts_table(void)
{
    uint32_t bytes_to_write = sizeof(_recs_counts_cache);
    uint32_t bytes_written;

    hb_stor_set_rw_pos(&_cfg_file, DEFAULT_RECS_COUNTS_OFFSET);
    bytes_written = hb_stor_write_buf(&_cfg_file, &_recs_counts_cache[0], bytes_to_write);

    // AK TODO: calculate crc16 for this section? How will it affect performance?
    hb_stor_sync_file(&_cfg_file);
    return (bytes_written == bytes_to_write) ? HB_SUCCESS : HB_FAIL;
}
/* --------------------------------------------------------------- */

static uint32_t _read_min_valid_rec_idx_table(void)
{
    uint32_t bytes_to_read = sizeof(_min_valid_rec_idx_cache);
    uint32_t bytes_read;

    hb_stor_set_rw_pos(&_cfg_file, DEFAULT_MIN_VALID_REC_IDX_OFFSET);
    bytes_read = hb_stor_read_buf(&_cfg_file, &_min_valid_rec_idx_cache[0], bytes_to_read);

    // AK TODO: check crc16 for this section? How will it affect performance?

    return (bytes_read == bytes_to_read) ? HB_SUCCESS : HB_FAIL;
}
/* --------------------------------------------------------------- */


static uint32_t _write_min_valid_rec_idx_table(void)
{
    uint32_t bytes_to_write = sizeof(_min_valid_rec_idx_cache);
    uint32_t bytes_written;

    hb_stor_set_rw_pos(&_cfg_file, DEFAULT_MIN_VALID_REC_IDX_OFFSET);
    bytes_written = hb_stor_write_buf(&_cfg_file, &_min_valid_rec_idx_cache[0], bytes_to_write);

    // AK TODO: calculate crc16 for this section? How will it affect performance?
    hb_stor_sync_file(&_cfg_file);
    return (bytes_written == bytes_to_write) ? HB_SUCCESS : HB_FAIL;
}
/* --------------------------------------------------------------- */


static uint32_t _init_file_tables(void)
{
    hal_memset(&_recs_counts_cache[0], 0, sizeof(_recs_counts_cache));
    hal_memset(&_min_valid_rec_idx_cache[0], 0, sizeof(_min_valid_rec_idx_cache));
    _write_rec_counts_table();
    _write_min_valid_rec_idx_table();

    // AK TODO: complete implementations for all other file tables
    return HB_SUCCESS;
}
/* --------------------------------------------------------------- */


/* returns HB_FAIL if metadata can't be used (either read error or crc16 failure) */
static uint32_t _check_metadata_crc(DATA_LOG_METADATA* p_metadata)
{
    return (p_metadata->crc16 == crc16((uint8_t*)p_metadata, sizeof(p_metadata->data_buf)) ? HB_SUCCESS : HB_FAIL);
}
/* --------------------------------------------------------------- */


/* returns HB_FAIL if metadata can't be used (either read error or crc16 failure) */
static void _calc_metadata_crc(DATA_LOG_METADATA* p_metadata)
{
    p_metadata->crc16 = crc16((uint8_t*)p_metadata, sizeof(p_metadata->data_buf));
}
/* --------------------------------------------------------------- */


/* returns HB_FAIL if metadata can't be read */
static uint32_t _read_metadata(uint32_t file_id,  DATA_LOG_METADATA* p_metadata)
{
    uint32_t bytes_to_read = sizeof(DATA_LOG_METADATA);
    uint32_t bytes_read;
    uint32_t offset = DEFAULT_METADATA_OFFSET + (SECTION_SIZE * file_id);

    hb_stor_set_rw_pos(&_cfg_file, offset);
    bytes_read = hb_stor_read_buf(&_cfg_file, p_metadata, bytes_to_read);

    return (bytes_read == bytes_to_read) ? HB_SUCCESS : HB_FAIL;
}
/* --------------------------------------------------------------- */


static uint32_t _write_metadata(uint32_t file_id, void* buf, uint32_t len)
{
    DATA_LOG_METADATA metadata;

    if (len <= sizeof(metadata.data_buf))
    {
        uint32_t bytes_to_write = sizeof(DATA_LOG_METADATA);
        uint32_t bytes_written;
        uint32_t offset = DEFAULT_METADATA_OFFSET + (SECTION_SIZE * file_id);

        hal_memcpy(&metadata.data_buf[0], buf, len);
        hal_memset(&metadata.data_buf[len], 0, sizeof(metadata.data_buf) - len);

        _calc_metadata_crc(&metadata);

        hb_stor_set_rw_pos(&_cfg_file, offset);
        bytes_written = hb_stor_write_buf(&_cfg_file, &metadata, bytes_to_write);

        return (bytes_written == bytes_to_write) ? HB_SUCCESS : HB_FAIL;
    }

    return HB_FAIL;
}
/* --------------------------------------------------------------- */


static uint32_t _fill_in_data_log_header(DATA_LOG_HEADER* pHdr)
{
    pHdr->version_major = DATA_LOG_VERSION_MAJOR;
    pHdr->version_minor = DATA_LOG_VERSION_MINOR;
    pHdr->max_data_logs = TOTAL_DATA_LOGS;
    pHdr->size          = sizeof(DATA_LOG_HEADER);

    pHdr->size_of_metadata = sizeof(DATA_LOG_METADATA);
    pHdr->number_of_tables = NUMBER_OF_PER_FILE_TABLES;

    hal_memset(&pHdr->reserved2, 0, sizeof(pHdr->reserved2));

    return HB_SUCCESS;
}
/* --------------------------------------------------------------- */


static uint32_t _write_data_log_header(DATA_LOG_HEADER* pHdr)
{
    uint32_t bytes_to_write = sizeof(*pHdr);
    uint32_t bytes_written;

    _fill_in_data_log_header(pHdr);
    pHdr->crc16 = crc16((uint8_t*)pHdr, sizeof(*pHdr) - 2);

    hb_stor_set_rw_pos(&_cfg_file, 0);
    bytes_written = hb_stor_write_buf(&_cfg_file, pHdr, bytes_to_write);

    if (bytes_written != bytes_to_write) {
        DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED writing data log header\n");)
        return HB_FAIL;
    }
    hb_stor_sync_file(&_cfg_file);
    return HB_SUCCESS;
}
/* --------------------------------------------------------------- */


static uint32_t _read_data_log_header(DATA_LOG_HEADER* pHdr)
{
    uint32_t bytes_to_read = sizeof(*pHdr);
    uint32_t bytes_read;

    hb_stor_set_rw_pos(&_cfg_file, 0);
    bytes_read = hb_stor_read_buf(&_cfg_file, pHdr, bytes_to_read);

    if (bytes_read != bytes_to_read) {
        DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED reading data log header\n");)
        return HB_FAIL;
    }

    if (pHdr->crc16 != crc16((uint8_t*)pHdr, sizeof(*pHdr) - 2)) {
        DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG - incorrect log header CRC\n");)
        return HB_FAIL;
    }

    return HB_SUCCESS;
}
/* --------------------------------------------------------------- */


/* returns HB_SUCCESS if we can safely migrate to the specified format */
static uint32_t _check_data_log_format(DATA_LOG_HEADER* pHdr)
{
    if ((pHdr->version_major == DATA_LOG_VERSION_MAJOR) &&
    (pHdr->version_minor == DATA_LOG_VERSION_MINOR) &&
    (pHdr->max_data_logs == TOTAL_DATA_LOGS) &&
    (pHdr->size == sizeof(DATA_LOG_HEADER)) &&
    (pHdr->size_of_metadata == sizeof(DATA_LOG_METADATA)) &&
    (pHdr->number_of_tables == NUMBER_OF_PER_FILE_TABLES)
    )
    {
        return HB_SUCCESS;
    }

    return HB_FAIL;
}
/* --------------------------------------------------------------- */


#if defined(HB_ITS_ONLY) || (defined(HB_WIN32) && defined(HB_UNIT_TESTS))
void hb_data_log_remove_cfg(void)
{
    char file_fname[MAX_FNAME_LEN];

    _compose_fname(FNAME_PREFIX, DATA_LOGS_CFG_FILE_NAME, CFG_FILE_EXT, file_fname);

    hb_stor_delete_file(file_fname);
}
#endif // HB_ITS_ONLY
/* --------------------------------------------------------------- */


uint32_t hb_data_log_init(void)
{
    uint32_t rc = HB_SUCCESS;

    char file_fname[MAX_FNAME_LEN];

    DATA_LOG_HEADER hdr;

    hal_memset(&_recs_counts_cache[0], 0, sizeof(_recs_counts_cache));

    _compose_fname(FNAME_PREFIX, DATA_LOGS_CFG_FILE_NAME, CFG_FILE_EXT, file_fname);

    if (hb_stor_open_file(&_cfg_file, file_fname) != HB_SUCCESS)
    {
        DBG_STOR2(HBPRINT("** WARN: HB_DATA_LOG - can't open .cfg file. Re-creating...\n");)

        if (hb_stor_create_log_file(&_cfg_file, file_fname, CFG_FILE_SIZE) != HB_SUCCESS) {
            DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED opening .cfg file\n");)
            rc = HB_FAIL;
        }

        if (_write_data_log_header(&hdr) != HB_SUCCESS) {
            rc = HB_FAIL;
        }
        _init_file_tables();
    }
    else
    {
        if (_read_data_log_header(&hdr) != HB_SUCCESS)
        {
            rc = HB_FAIL;
        }
        else
        {
            if (_check_data_log_format(&hdr) != HB_SUCCESS)
            {
                DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG - unexpected format of .cfg file. Re-creating...\n");)

                // AK TODO: currently we just updating DATA_LOG_INFO, but in future we may
                //          need to do something else (for example, convert it...)

                if (_write_data_log_header(&hdr) != HB_SUCCESS) {
                    rc = HB_FAIL;
                }
            }

            if (_read_rec_counts_table() != HB_SUCCESS) {
                DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED reading recs size\n");)
                rc = HB_FAIL;
            }
            if (_read_min_valid_rec_idx_table() != HB_SUCCESS) {
                DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED reading min valid idx table\n");)
                rc = HB_FAIL;
            }
        }

        if (rc != DF_STATUS_OK) {
            hb_stor_close_file(&_cfg_file);
        }
    }

    return rc;
}


/* --------------------------------------------------------------- */

void hb_data_log_uninit(void)
{
    hb_stor_sync_file(&_cfg_file);
    hb_stor_close_file(&_cfg_file);
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_save_metadata(uint32_t file_id, void* p_buf, uint32_t size)
{
    if (_write_metadata(file_id, p_buf, size) != HB_SUCCESS) {
        DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED writing metadata\n");)
        return HB_FAIL;
    }

    return HB_SUCCESS;
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_read_metadata(uint32_t file_id, void* p_buf, uint32_t size)
{
    DATA_LOG_METADATA metadata;

    if (_read_metadata(file_id, &metadata) != HB_SUCCESS)
    {
        DBG_ERR(HBPRINT("### ERR: HB_DATA_LOG - FAILED reading metadata\n");)
        return HB_FAIL;
    }
    else
    {
        if (_check_metadata_crc(&metadata) != HB_SUCCESS)
        {
            DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG - incorrect metadata CRC\n");)
            return HB_FAIL;
        }
    }

    if (size > sizeof(DATA_LOG_METADATA)) {
        size = sizeof(DATA_LOG_METADATA);
        DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG - too big size of requested metadata\n");)
    }

    hal_memcpy(p_buf, &metadata, size);

    return HB_SUCCESS;
}
/* --------------------------------------------------------------- */


#define LOG_FILE_PIECES 10

static uint32_t _compose_extension_and_offset_from_record_number(HB_DATA_LOG* pLog, uint32_t rnum, int32_t *filenum, char ext[5]) {
	const uint32_t num_records_per_file = pLog->max_recs / LOG_FILE_PIECES;
	uint32_t foffset = (rnum % num_records_per_file) * pLog->rec_size;
	uint32_t fnum = (rnum / num_records_per_file) % 1000;
	snprintf(ext, 5, ".%03i", (int)fnum);
    if (filenum) {
        *filenum = fnum;
    }        
	return foffset;
}

static void _close_read_handle(HB_DATA_LOG* pLog) {
    if (pLog->read_file_index != -1) {
        hb_stor_close_file(&pLog->read_file);
        pLog->read_file_index = -1;
    }
}

uint32_t hb_data_log_create(HB_DATA_LOG* pLog)
{
	//HBPRINT("%s\n", __FUNCTION__);

    char dat_file_fname[MAX_FNAME_LEN];
	
	char current_log_file_ext[5];

	uint32_t rec_offset = _compose_extension_and_offset_from_record_number(pLog, 0, NULL, current_log_file_ext);
    (void)rec_offset;
    //uint32_t file_size = pLog->rec_size * pLog->max_recs;

    _compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);

    /* reinsurance */
    hb_stor_delete_file(dat_file_fname);

    //if (hb_stor_create_log_file(&pLog->dat_file, dat_file_fname, file_size) != HB_SUCCESS) {
	if (hb_stor_create_file(&pLog->dat_file, dat_file_fname) != HB_SUCCESS) {
        return DF_STATUS_FAIL;
    }

    _recs_counts_cache[pLog->file_id] = 0;
    pLog->dat_file.recs_count = 0;                               // backward compatibility
    _min_valid_rec_idx_cache[pLog->file_id] = 0;

    if ((_write_rec_counts_table() == HB_SUCCESS) &&
        (_write_min_valid_rec_idx_table() == HB_SUCCESS)) {
        return DF_STATUS_OK;
    }

    return DF_STATUS_FAIL;
}
/* --------------------------------------------------------------- */

uint32_t hb_data_log_open(HB_DATA_LOG* pLog)

{
	//HBPRINT("%s\n", __FUNCTION__);

    uint32_t rc = DF_STATUS_FAIL;

    char dat_file_fname[MAX_FNAME_LEN];

	char current_log_file_ext[5];

	uint32_t rec_offset = _compose_extension_and_offset_from_record_number(pLog, _recs_counts_cache[pLog->file_id], NULL, current_log_file_ext);
	
	_compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);

    if (hb_stor_open_file(&pLog->dat_file, dat_file_fname) == HB_SUCCESS)
    {
        uint32_t recs_count = _recs_counts_cache[pLog->file_id];

        pLog->dat_file.recs_count = recs_count;                      // backward compatibility

        hb_stor_set_rw_pos(&pLog->dat_file, rec_offset);

        rc = DF_STATUS_OK;

        DBG_INIT(HBPRINT("%s contains %ld records\n", dat_file_fname, recs_count);)
    }

    return rc;
}
/* --------------------------------------------------------------- */


void hb_data_log_close(HB_DATA_LOG* pLog)
{
	//HBPRINT("%s\n", __FUNCTION__);

    _close_read_handle(pLog);
    hb_stor_close_file(&pLog->dat_file);
	pLog->dat_file.handle = NULL;
}
/* --------------------------------------------------------------- */


void hb_data_log_delete(HB_DATA_LOG* pLog)
{
	//HBPRINT("%s\n", __FUNCTION__);

    char dat_file_fname[MAX_FNAME_LEN];

	const uint32_t num_records_per_file = pLog->max_recs / LOG_FILE_PIECES;

	char current_log_file_ext[5];

	for (int i = 0; i < LOG_FILE_PIECES + 2; i++) {
		uint32_t rec_offset = _compose_extension_and_offset_from_record_number(pLog, _recs_counts_cache[pLog->file_id]-num_records_per_file*i, NULL, current_log_file_ext);
        (void)rec_offset;

		_compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);

		hb_stor_delete_file(dat_file_fname);
	}

    _recs_counts_cache[pLog->file_id] = 0;
    _min_valid_rec_idx_cache[pLog->file_id] = 0;
    _write_rec_counts_table();         // for consistency
    _write_min_valid_rec_idx_table();
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_read_record(HB_DATA_LOG* pLog, uint32_t rec_id, void* dst_buf)
{
	//HBPRINT("%s\n", __FUNCTION__);

    uint32_t rc = DF_STATUS_FAIL;
    //uint32_t offset = (rec_id % pLog->max_recs) * pLog->rec_size;
    uint32_t bytes_read;

	char dat_file_fname[MAX_FNAME_LEN];

	char current_log_file_ext[5];
	f_handle_t *phandle;//, read_handle = { NULL, 0, 0 };

	const uint32_t num_records_per_file = pLog->max_recs / LOG_FILE_PIECES;
	//bool need_to_close = true;

	if (rec_id >= _recs_counts_cache[pLog->file_id]) {
		return DF_STATUS_FAIL;
	}
    
    int32_t filenum = -1;
	uint32_t rec_offset = _compose_extension_and_offset_from_record_number(pLog, rec_id, &filenum, current_log_file_ext);

	if ((rec_id / num_records_per_file == _recs_counts_cache[pLog->file_id] / num_records_per_file)) {
		phandle = &pLog->dat_file;
		//need_to_close = false;
	}
	else {
        if (pLog->read_file_index != filenum) {
            _close_read_handle(pLog);
		    _compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);
		    if (hb_stor_open_file(&pLog->read_file, dat_file_fname) != HB_SUCCESS) {
    			//HBPRINT("Errno: %i\n", errno);
	    		return DF_STATUS_FAIL;
		    }
            pLog->read_file_index = filenum;
        }            
        phandle = &pLog->read_file;
	}

	if (HB_SUCCESS == hb_stor_set_rw_pos(phandle, rec_offset)) {
        //memset(dst_buf, 0x55, pLog->rec_size);
		bytes_read = hb_stor_read_buf(phandle, dst_buf, pLog->rec_size);

		if (bytes_read == pLog->rec_size) {
			rc = DF_STATUS_OK;
		}
	}

	//if (need_to_close) {
	//	hb_stor_close_file(&read_handle);
	//}

    return rc;
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_save_record(HB_DATA_LOG* pLog, uint32_t rec_id, void* src_buf)
{
	//HBPRINT("%s\n", __FUNCTION__);

	uint32_t rc = DF_STATUS_FAIL;
	//uint32_t offset = (rec_id % pLog->max_recs) * pLog->rec_size;
	uint32_t bytes_written;

	char dat_file_fname[MAX_FNAME_LEN];

	char current_log_file_ext[5];
	f_handle_t write_handle = { NULL, 0, 0 };

	const uint32_t num_records_per_file = pLog->max_recs / LOG_FILE_PIECES;
	bool need_to_close = true;

	hal_feed_the_dog();

	if (rec_id > _recs_counts_cache[pLog->file_id]) {
		// AK TODO: what should we do if the caller wants to save an out of order record?
		DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG [%ld] - can't save data record\n", pLog->file_id);)
		return DF_STATUS_FAIL;
	}

	uint32_t rec_offset = _compose_extension_and_offset_from_record_number(pLog, rec_id, NULL, current_log_file_ext);

	if ((rec_id / num_records_per_file == _recs_counts_cache[pLog->file_id] / num_records_per_file)) {
		if (pLog->dat_file.handle == NULL) {
			return DF_STATUS_FAIL;
		}
		write_handle = pLog->dat_file;
		need_to_close = false;
	}
	else {
        _close_read_handle(pLog);
		_compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);
		if (hb_stor_open_file(&write_handle, dat_file_fname) != HB_SUCCESS) {
			return DF_STATUS_FAIL;
		}
	}

	if (HB_SUCCESS == hb_stor_set_rw_pos(&write_handle, rec_offset)) {
        if (rec_id == _recs_counts_cache[pLog->file_id]) {
            hb_stor_trunc_file(&write_handle, rec_offset);
        }            
		bytes_written = hb_stor_write_buf(&write_handle, src_buf, pLog->rec_size);

		if (bytes_written == pLog->rec_size) {
			rc = DF_STATUS_OK;
		}

	}

	if (need_to_close) {
		hb_stor_close_file(&write_handle);
	}

	if ((rc == DF_STATUS_OK) && (rec_id == _recs_counts_cache[pLog->file_id]))
	{
        if (_min_valid_rec_idx_cache[pLog->file_id] == _recs_counts_cache[pLog->file_id] - pLog->max_recs) {
            _min_valid_rec_idx_cache[pLog->file_id]++;
            _write_min_valid_rec_idx_table();// ToDO: check error
        }

		_recs_counts_cache[pLog->file_id]++;
		pLog->dat_file.recs_count = _recs_counts_cache[pLog->file_id];     // backward compatibility

		if (_write_rec_counts_table() != HB_SUCCESS) {
			// AK TODO: what should we do if saving the number of records fails?
			DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG [%ld] - can't add data record\n", pLog->file_id);)
				return DF_STATUS_FAIL;
		}

		if (rec_id / num_records_per_file != _recs_counts_cache[pLog->file_id] / num_records_per_file) {
			hb_stor_close_file(&pLog->dat_file);
			uint32_t rec_offset_ = _compose_extension_and_offset_from_record_number(pLog, _recs_counts_cache[pLog->file_id], NULL, current_log_file_ext);
            (void)rec_offset_;
			_compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);
			/* reinsurance */
			hb_stor_delete_file(dat_file_fname);
			if (hb_stor_create_file(&pLog->dat_file, dat_file_fname) != HB_SUCCESS) {
				return DF_STATUS_FAIL;
			}

			int32_t last_record_number = _recs_counts_cache[pLog->file_id] - (num_records_per_file * (LOG_FILE_PIECES + 2));
			if (last_record_number >= 0) {
				rec_offset_ = _compose_extension_and_offset_from_record_number(pLog, last_record_number, NULL, current_log_file_ext);
				assert(rec_offset_ == 0);
				_compose_fname(FNAME_PREFIX, pLog->log_name, current_log_file_ext, dat_file_fname);
                _close_read_handle(pLog);
				hb_stor_delete_file(dat_file_fname);
                _min_valid_rec_idx_cache[pLog->file_id] = max(last_record_number + num_records_per_file, _min_valid_rec_idx_cache[pLog->file_id]);
                _write_min_valid_rec_idx_table();
			}
		}
	}


	return rc;
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_get_num_records(HB_DATA_LOG* pLog)
{
	//HBPRINT("%s\n", __FUNCTION__);
	return _recs_counts_cache[pLog->file_id];
}
/* --------------------------------------------------------------- */

uint32_t hb_data_log_get_first_rec_idx(HB_DATA_LOG* pLog)
{
    //HBPRINT("%s\n", __FUNCTION__);
    //if (_recs_counts_cache[pLog->file_id] > pLog->max_recs) {
    //    return _recs_counts_cache[pLog->file_id] - pLog->max_recs;
    //} else {
    //    return 0;
    //}
    return _min_valid_rec_idx_cache[pLog->file_id];
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_get_size_of_data(HB_DATA_LOG* pLog)
{
	//HBPRINT("%s\n", __FUNCTION__);
	return min(_recs_counts_cache[pLog->file_id], pLog->max_recs) * pLog->rec_size;
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_add_record(HB_DATA_LOG* pLog, void* psrc_buf)
{
	//HBPRINT("%s\n", __FUNCTION__);
	return hb_data_log_save_record(pLog, _recs_counts_cache[pLog->file_id], psrc_buf);
}
/* --------------------------------------------------------------- */


uint32_t hb_data_log_remove_last_record(HB_DATA_LOG* pLog)
{
	//HBPRINT("%s\n", __FUNCTION__);
	if (_recs_counts_cache[pLog->file_id] > _min_valid_rec_idx_cache[pLog->file_id])
    {
		char current_log_file_ext[5];

        _close_read_handle(pLog);
		uint32_t rec_offset = _compose_extension_and_offset_from_record_number(pLog, _recs_counts_cache[pLog->file_id], NULL, current_log_file_ext);
		if (rec_offset == 0) {
			// Need to focus on previous file
			hb_data_log_close(pLog);
		}

		_recs_counts_cache[pLog->file_id]--;

		if (rec_offset == 0) {
			// Need to focus on previous file
			hb_data_log_open(pLog);
		}
		pLog->dat_file.recs_count = _recs_counts_cache[pLog->file_id];         // backward compatibility

        if (_write_rec_counts_table() != HB_SUCCESS) {
            // AK TODO: what should we do if saving the number of records fails?
            DBG_WARN(HBPRINT("** WARN: HB_DATA_LOG [%ld] - can't remove last record\n", pLog->file_id);)
            return DF_STATUS_FAIL;
        }
    }

    return DF_STATUS_OK;
}
/* --------------------------------------------------------------- */


int32_t hb_data_log_read_data(HB_DATA_LOG* pLog, uint32_t offset, uint32_t bytes_to_read, void* pdst_buf, uint32_t* p_bytes_read)
{
	//HBPRINT("%s\n", __FUNCTION__);
	int32_t  rc = HB_ERROR;
	uint8_t *pdstbuf = pdst_buf;
	*p_bytes_read = 0;
    uint8_t *buf = NULL;

    // AK TODO: DO NOT CLOSE ONCE OPENED FILE HERE(!) and make hb_data_log_open() first
    //          check if the file is already opened. This can save us a lot of time when
    //          this function is called from IO when a log file is read by an external app.

    do
    {
		if ((pdst_buf == NULL) || (p_bytes_read == NULL)) {
			break;
		}

		int32_t current_index_of_zero_offset_record  =  _recs_counts_cache[pLog->file_id] / pLog->max_recs      * pLog->max_recs;
		//int32_t previous_index_of_zero_offset_record = (_recs_counts_cache[pLog->file_id] / pLog->max_recs - 1) * pLog->max_recs; // may be negative
		int32_t record_num_cyclic = offset / pLog->rec_size % pLog->max_recs;
		int32_t remainder_offset_cyclic = offset % pLog->rec_size;
        //uint8_t buf[pLog->rec_size];
        buf = alloca(pLog->rec_size);
		bool need_to_close = false;

		if (!pLog->dat_file.handle) {
			if (DF_STATUS_OK != hb_data_log_open(pLog)) {
				return rc;
			}
			need_to_close = true;
		}
		while (bytes_to_read) {
			if (current_index_of_zero_offset_record + record_num_cyclic >= (int32_t)_recs_counts_cache[pLog->file_id]) {
				record_num_cyclic -= pLog->max_recs;
			}
			//uint8_t buf[pLog->rec_size];
			rc = hb_data_log_read_record(pLog, current_index_of_zero_offset_record + record_num_cyclic, buf);
			if (DF_STATUS_OK != rc) {
				break;
			}

			int32_t toread = min(pLog->rec_size - remainder_offset_cyclic, bytes_to_read);
			memcpy(pdstbuf, &buf[remainder_offset_cyclic], toread);
			remainder_offset_cyclic = 0;
			record_num_cyclic++;
			bytes_to_read -= toread;
			pdstbuf += toread;
			*p_bytes_read += toread;
		}

		if (need_to_close) {
			hb_data_log_close(pLog);
		}

    } while (0);
    
    //if (buf) {
    //    free(buf);
    //}

    return rc;
}
/* ==================================================================== */
