/*
 * hb_data_log.h
 *
 * Created: 21.08.2019 10:19:18
 *  Author: 
 */ 


#ifndef HB_DATA_LOG_H_
#define HB_DATA_LOG_H_

#include "os_dependency.h"          /* platform dependent definitions and includes */
#include "hb_storage.h"

/*************************************************************************
*  HB_DATA_LOG object for working with various data records
*************************************************************************/

// NOTE: HB_DATA_LOG object represents a couple of items, one of which represents
//       a PREALLOCATED cyclical buffer for storing the data records and the other
//       one stores the *TOTAL* number of records whenever put into that buffer.
//       So, the *total number* may become bigger than the number of records that
//       can fit into the buffer. With this approach the earliest records may and
//       will be overwritten by later records, but if the size of the buffer is big
//       enough, we're just not interested in those earlier records.
//       In current implementation the both mentioned items are physical files (of
//       a constant size). Later on we can use some dedicated memory buffers in a
//       flash memory for storing data and a specialized data structure for storing
//       the available number of records in each such buffer.

#define DATA_FILE_ID_UNKNOWN           0    // reserved for module internal use


#define MAX_DATA_FILE_ID               16
#define TOTAL_DATA_LOGS                32   // AK 2016-10-08: DO NOT EVER CHANGE THIS!!!

typedef struct data_log_
{
    const char*    log_name;   // bare name w/o prefix (filepath) and suffix (file extension), max 8 chars
    const uint32_t file_id;    // unique file id (see definitions for DATA_FILE_ID_xxx above)
    const uint32_t rec_size;   // size of the records, stored in the .dat file (must be inited)
    const uint32_t max_recs;   // max records in the .dat file
    int32_t read_file_index;         //
    f_handle_t dat_file;       // for storing data records
    f_handle_t read_file;      // for reading data records

} HB_DATA_LOG;

#define DF_STATUS_OK    HB_SUCCESS
#define DF_STATUS_FAIL  HB_FAIL

uint32_t hb_data_log_init(void);
void     hb_data_log_uninit(void);
void     hb_data_log_remove_cfg(void);

uint32_t hb_data_log_save_metadata(uint32_t file_id, void* p_buf, uint32_t size);
uint32_t hb_data_log_read_metadata(uint32_t file_id, void* p_buf, uint32_t size);

uint32_t hb_data_log_create(HB_DATA_LOG* pLog);
uint32_t hb_data_log_open(HB_DATA_LOG* pLog);
void     hb_data_log_close(HB_DATA_LOG* pLog);
void     hb_data_log_delete(HB_DATA_LOG* pLog);

uint32_t hb_data_log_read_record(HB_DATA_LOG* pLog, uint32_t rec_id, void* dat_buf);
uint32_t hb_data_log_save_record(HB_DATA_LOG* pLog, uint32_t rec_id, void* src_buf);

/* works for already opened file */
uint32_t hb_data_log_get_num_records(HB_DATA_LOG* pLog);
uint32_t hb_data_log_get_first_rec_idx(HB_DATA_LOG* pLog);
uint32_t hb_data_log_get_size_of_data(HB_DATA_LOG* pLog);

/* saves a new record right after the last existing record */
uint32_t hb_data_log_add_record(HB_DATA_LOG* pLog, void* src_buf);

uint32_t hb_data_log_remove_last_record(HB_DATA_LOG* pLog);

/* functions for retrieving data from possibly no opened DATA_FILEs */
int32_t hb_data_log_read_data(HB_DATA_LOG* pLog, uint32_t offset, uint32_t bytes_to_read, void *dest_buf, uint32_t *p_bytes_read);
/* ==================================================================== */
#endif /* HB_DATA_LOG_H_ */
