/**
 * \file
 *
 * \brief Implementation of the  Core FAT FS related test routines
 *
 */

#include "os_dependency.h"          /* platform dependent definitions and includes */

//#include "Core/hb_core.h"
//#include "Core/hb_DateTime.h"
#include "Core/hb_hal.h"
#include "hb_storage.h"
#include "Utils/hb_utils.h"       // crc16()

#if defined(HB_ATMEL_ARM) || defined(EMULATE_TARGET_FS)
//#include <asf.h>
#include <string.h>

#if !defined(EMULATE_TARGET_FS)
#include "ff.h"
#endif
#include "lfs.h"


// AK TODO: enable "fast seek" feature

/* ******************************************************************
 * Local types,  definitions and variables
 * *****************************************************************/

#if defined(FAT_FS)

#define HB_FIL      FIL
#define HB_DIR      DIR

#define HB_FS_READ              (FA_READ)
#define HB_FS_WRITE             (FA_WRITE)
#define HB_FS_OPEN_EXISTING     (FA_OPEN_EXISTING)
#define HB_FS_OPEN_ALWAYS       (FA_OPEN_ALWAYS)
#define HB_FS_CREATE_ALWAYS     (FA_CREATE_ALWAYS)
#define HB_FS_CREATE_NEW        (FA_CREATE_NEW)

static FATFS _fs;

typedef struct
{
    BYTE    memory_lun_id;
    FATFS   *pfs;
} hb_fs_context_t;

static hb_fs_context_t hb_fs_context =
{
    .pfs = &_fs
};
#endif // FAT_FS


#if defined(LITTLE_FS)

#define HB_FIL      lfs_file_t
#define HB_DIR      lfs_dir_t

#define HB_FS_READ              (LFS_O_RDONLY)
#define HB_FS_WRITE             (LFS_O_WRONLY)
#define HB_FS_OPEN_EXISTING     (0)
#define HB_FS_OPEN_ALWAYS       (LFS_O_CREAT)
#define HB_FS_CREATE_ALWAYS     (LFS_O_CREAT)
#define HB_FS_CREATE_NEW        (LFS_O_TRUNC)

static lfs_t _lfs;
static struct lfs_config _cfg;

typedef struct
{
    struct lfs_config *pcfg;
    lfs_t *pfs;
    Bool mounted;
} hb_fs_context_t;

static /*const*/ hb_fs_context_t hb_fs_context =
{
    .pcfg = &_cfg,
    .pfs = &_lfs,
    .mounted = HB_FALSE,
};
#endif // LITTLE_FS


typedef struct
{
    uint32_t (*init_cfg)(void *pfs_context);
    uint32_t (*mount)   (void *pfs_context);
    uint32_t (*unmount) (void *pfs_context);
    uint32_t (*open)    (void *pfs_context, HB_FIL *pfile, const char *pfname, uint32_t mode);
    uint32_t (*format)  (void *pfs_context);
    uint32_t (*seek)    (void *pfs_context, HB_FIL *pfile, uint32_t offset);
    uint32_t (*write)   (void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbw);
    uint32_t (*read)    (void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbr);
    uint32_t (*clear)   (void *pfs_context, HB_FIL *pfile);
    uint32_t (*close)   (void *pfs_context, HB_FIL *pfile);
    uint32_t (*delete)  (void *pfs_context, const char* pfname);
    uint32_t (*tell)    (void *pfs_context, HB_FIL *pfile, uint32_t *ps);
    uint32_t (*sync)    (void *pfs_context, HB_FIL *pfile);
    uint32_t (*get_size)(void *pfs_context, HB_FIL *pfile, uint32_t *psize);
    uint32_t (*trunc)   (void *pfs_context, HB_FIL *pfile, const uint32_t size);
    uint32_t (*stat)    (void *pfs_context, const char *ppath, file_info_t *pfno);
    uint32_t (*rename)  (void *pfs_context, const char *ppath_old, const char *ppath_new);

} hb_fs_t;


#ifdef HB_FUNCTIONAL_FW
#define MAX_FILES           128       // NOTE: 512+ bytes per file handle!!!
#else
#define MAX_FILES           _FS_SHARE // NOTE: 512+ bytes per file handle!!!
#endif
#define FR_NO_FILE_OBJECTS   128       // our custom error code

static struct
{
    HB_FIL   file_obj[MAX_FILES];
    uint16_t in_use[MAX_FILES];
    uint16_t count;
} _file_obj_pool;


static int32_t _error_code;
/* =============================================================== */


/* ******************************************************************
 * Local functions
 * *****************************************************************/
static uint32_t _fs_init_cfg(void *pfs_context);
static uint32_t _fs_mount(void *pfs_context);
static uint32_t _fs_unmount(void *pfs_context);
static uint32_t _fs_open(void *pfs_context, HB_FIL *pfile, const char* pfname, uint32_t mode);
static uint32_t _fs_format(void *pfs_context);
static uint32_t _fs_seek(void *pfs_context, HB_FIL *pfile, const uint32_t offset);
static uint32_t _fs_write(void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbw);
static uint32_t _fs_read(void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbr);
static uint32_t _fs_clear(void *pfs_context, HB_FIL *pfile);
static uint32_t _fs_close(void *pfs_context, HB_FIL *pfile);
static uint32_t _fs_delete(void *pfs_context, const char* pfname);
static uint32_t _fs_tell(void *pfs_context, HB_FIL *pfile, uint32_t *ps);
static uint32_t _fs_sync(void *pfs_context, HB_FIL *pfile);
static uint32_t _fs_get_size(void *pfs_context, HB_FIL *pfile, uint32_t *psize);
static uint32_t _fs_trunc_file(void *pfs_context, HB_FIL *pfile, const uint32_t size);
static uint32_t _fs_stat(void *pfs_context, const char *ppath, file_info_t *pfno);
static uint32_t _fs_rename(void *pfs_context, const char *ppath_old, const char *ppath_new);

static const hb_fs_t hb_fs =
{
    .init_cfg   = _fs_init_cfg,
    .mount      = _fs_mount,
    .open       = _fs_open,
    .unmount    = _fs_unmount,
    .format     = _fs_format,
    .seek       = _fs_seek,
    .write      = _fs_write,
    .read       = _fs_read,
    .clear      = _fs_clear,
    .close      = _fs_close,
    .delete     = _fs_delete,
    .tell       = _fs_tell,
    .sync       = _fs_sync,
    .get_size   = _fs_get_size,
    .trunc      = _fs_trunc_file,
    .stat       = _fs_stat,
    .rename     = _fs_rename,
};

static HB_FIL* _allocate_file_object(uint32_t* pidx);
static void _free_file_object(f_handle_t* phandle);
static uint32_t _open_file(f_handle_t* phandle, const char* fname, uint32_t mode);

/* =============================================================== */

static HB_FIL* _allocate_file_object(uint32_t* pidx)
{
    HB_FIL* p_file_obj = NULL;

    hal_EnterCriticalSection();
    if (_file_obj_pool.count < MAX_FILES)
    {
        uint32_t i;

        for (i = 0; i < MAX_FILES; i++)
        {
            if (_file_obj_pool.in_use[i] == HB_FALSE)
            {
                _file_obj_pool.count++;
                _file_obj_pool.in_use[i] = HB_TRUE;
                *pidx = i;

                p_file_obj = &_file_obj_pool.file_obj[i];
                break;
            }
        }
    }
    hal_LeaveCriticalSection();

#if !defined(HB_BASE_FW)
#if defined(HB_DEBUG) && !defined(EMULATE_TARGET_FS)
    if (p_file_obj == NULL) {
        hb_core_register_error(ERROR_EXCESSIVE_FILE_OBJECT_ALLOC);
    }
#endif // HB_DEBUG
#endif // HB_BASE_FW

    return p_file_obj;
}
/* =============================================================== */


static void _free_file_object(f_handle_t* phandle)
{
    if (phandle->idx < MAX_FILES) {
        hal_EnterCriticalSection();
        if (_file_obj_pool.in_use[phandle->idx] == HB_TRUE) {
            _file_obj_pool.in_use[phandle->idx] = HB_FALSE;
            _file_obj_pool.count--;
            hal_LeaveCriticalSection();
        } else {
            hal_LeaveCriticalSection();
#if !defined(HB_BASE_FW)
#if defined(HB_DEBUG) && !defined(EMULATE_TARGET_FS)
            hb_core_register_error(ERROR_EXCESSIVE_FILE_OBJECT_FREE);
            DBG_ERR(HBPRINT("### ERR: error freeing file object %ld\n", phandle->idx);)
#endif // HB_DEBUG
#endif // HB_BASE_FW
        }
    }
    return;

}
/* =============================================================== */


uint32_t _open_file(f_handle_t* phandle, const char* fname, uint32_t mode)
{
    uint32_t    idx;
    uint32_t    res = HB_FAIL;
    HB_FIL*     pfile = _allocate_file_object(&idx);

    if (pfile != NULL)
    {
        phandle->handle     = pfile;
        phandle->idx        = idx;
        phandle->recs_count = 0;

        res = hb_fs.open((void*)&hb_fs_context, pfile, fname, mode);

        if (res != HB_SUCCESS) {
            return res;
        }

        return HB_SUCCESS;
    }

    DBG_WARN(HBPRINT("** WARN: _open_file() FAILED: no more file objects\n\n");)
    return HB_FAIL;
}
/* =============================================================== */


uint32_t hb_storage_init(void)
{
    uint32_t res;
    DBG_STOR(HBPRINT("Initializing HB Storage...\n");)

    hal_memset(&_file_obj_pool, 0, sizeof(_file_obj_pool));

    if (hb_fs.init_cfg) {
        res = hb_fs.init_cfg(&hb_fs_context);
        if (HB_SUCCESS != res) {
            DBG_STOR2(HBPRINT("### ERR: hb_storage_init() - initializing fs context FAILED\n");)
            return HB_FAIL;
        }
    }

    /* Mount a disk volume */
    DBG_STOR(HBPRINT("Mounting disk volume...");)

    res = hb_fs.mount((void*)&hb_fs_context);

    if (HB_SUCCESS != res) {
        DBG_STOR2(HBPRINT("\n");)
        DBG_STOR2(HBPRINT("### ERR: hb_storage_init() - mounting disk volume FAILED\n");)
        return HB_FAIL;
    }

    DBG_STOR(HBPRINT("[OK]\n\n");)
    return HB_SUCCESS;
}
/* =============================================================== */


uint32_t hb_storage_uninit(void)
{
    DBG_STOR(HBPRINT("Uninitializing HB Storage...\n\n");)
    return hb_fs.unmount((void*)&hb_fs_context);
}
/* =============================================================== */


uint32_t hb_stor_format_drive(HB_BOOL init)
{
    uint32_t res;

    if (init) {
        hal_turn_on_SD_card();
        hb_storage_init();
    }
    res = hb_fs.format((void*)&hb_fs_context);
    if (res != HB_SUCCESS) {
        DBG_STOR2(HBPRINT("### ERR: hb_stor_format_drive() FAILED [%ld]\n", res);)
        return HB_FAIL;
    }

    return HB_SUCCESS;
}
/* =============================================================== */


uint32_t hb_stor_create_log_file(f_handle_t* phandle, const char* fname, uint32_t size)
{
    /* open an existing file or create a new one */
    if (hb_stor_create_file(phandle, fname) != HB_SUCCESS) {
        return HB_FAIL;
    }

    /* clear the file if it has been open */
    //hb_stor_clear_file(phandle);               // this slows things down for existing files

    /* preallocate the required amount of space to avoid this overhead in future */
    if (hb_stor_set_rw_pos(phandle, size) != HB_SUCCESS) {
        DBG_WARN(HBPRINT("** WARN: not enough space for log file '%s'\n", fname);)
        return HB_FAIL;
    }

    /* return to the beginning of the file */
    hb_stor_set_rw_pos(phandle, 0);

    DBG_STOR(HBPRINT("    total reserved: %ld bytes\n", hb_stor_get_file_size(phandle));)

    // AK TODO: should we fill the newly created file with zeroes? Sounds good...

    //f_sync(phandle->handle);

    return HB_SUCCESS;
}
/* =============================================================== */


uint32_t hb_stor_create_file(f_handle_t* phandle, const char* fname)
{
    uint32_t res;

    DBG_STOR(HBPRINT("Creating file: %s...", fname);)

    res = _open_file(phandle, fname, HB_FS_OPEN_ALWAYS | HB_FS_READ | HB_FS_WRITE);

    if (res != HB_SUCCESS) {
        DBG_ERR(HBPRINT("\n### ERR: hb_stor_create_file() failed with code %ld\n", _error_code);)
        _free_file_object(phandle);
        //hb_stor_close_file(phandle);
        return HB_FAIL;
    }

    DBG_STOR(HBPRINT("   -   [OK]\r\n");)

    return HB_SUCCESS;
}
/* =============================================================== */


uint32_t hb_stor_open_file(f_handle_t* phandle, const char* pfname)
{
    uint32_t res;

    DBG_STOR(HBPRINT("Opening file: %s...", pfname);)

    res = _open_file(phandle, pfname, HB_FS_OPEN_EXISTING | HB_FS_READ | HB_FS_WRITE);

    if (res != HB_SUCCESS) {
        DBG_WARN(
            DBG_STOR2(HBPRINT("\nWARN: hb_stor_open_file() failed with code %ld\n", res););
        )
        _free_file_object(phandle);
        //hb_stor_close_file(phandle);
        return res;
    }

    DBG_STOR(HBPRINT("   -   [OK]\n");)

    return HB_SUCCESS;
}
/* =============================================================== */

uint32_t hb_stor_seek_file(f_handle_t* phandle, const uint32_t offset)
{
    return hb_fs.seek((void*)&hb_fs_context, phandle->handle, offset);
}
/* =============================================================== */

uint32_t hb_stor_write_file(f_handle_t* phandle, uint8_t* pbuf, uint32_t size, uint32_t* pbw)
{
    return hb_fs.write((void*)&hb_fs_context, phandle->handle, pbuf, size, pbw);
}
/* =============================================================== */

uint32_t hb_stor_read_file(f_handle_t* phandle, uint8_t* pbuf, uint32_t size, uint32_t* pbr)
{
    return hb_fs.read((void*)&hb_fs_context, phandle->handle, pbuf, size, pbr);
}
/* =============================================================== */

/* returns non-zero on success */
uint32_t hb_stor_file_is_opened(f_handle_t* phandle)
{
    return (phandle->handle != NULL);
}
/* =============================================================== */

uint32_t hb_stor_clear_file(f_handle_t* phandle)
{
    if (phandle->handle != NULL)    // for ability to safely call this function more than once
    {
        return hb_fs.clear((void*)&hb_fs_context, phandle->handle);
    }
    return HB_FAIL;
}
/* =============================================================== */

uint32_t hb_stor_close_file(f_handle_t* phandle)
{
    if (phandle->handle != NULL)       // to safely call this function more than once
    {
        uint32_t res = hb_fs.close((void*)&hb_fs_context, phandle->handle);

        _free_file_object(phandle);

        phandle->handle     = NULL;
        phandle->idx        = 0;
        phandle->recs_count = 0;

        return res;
    } else {
        return HB_FAIL;
    }
}
/* =============================================================== */

uint32_t hb_stor_delete_file(const char* pfname)
{
    return hb_fs.delete((void*)&hb_fs_context, pfname);
}
/* =============================================================== */

uint32_t hb_stor_tell_file(f_handle_t* phandle)
{
    uint32_t s = 0;
    if (hb_fs.tell((void*)&hb_fs_context, phandle->handle, &s) == HB_SUCCESS) {
        return s;
    }
    return HB_FAIL;

}
/* =============================================================== */

uint32_t hb_stor_sync_file(f_handle_t* phandle)
{
    return hb_fs.sync((void*)&hb_fs_context, phandle->handle);
}
/* =============================================================== */

int32_t hb_stor_size_file(f_handle_t* phandle)
{
    uint32_t size = 0;
    if (hb_fs.get_size((void*)&hb_fs_context, phandle->handle, &size) == HB_SUCCESS) {
        return size;
    }
    return HB_ERROR;
}
/* =============================================================== */

uint32_t hb_stor_wipe_file(f_handle_t* phandle, uint32_t offset, uint32_t size)
{
    uint8_t buf[256];
    uint32_t res;

    int32_t bytes_to_write = size;
    uint32_t bytes_written;

    hal_memset(buf, 0, sizeof(buf));

    res = hb_stor_seek_file(phandle, offset);

    if (res != 0) {
        return HB_FAIL;
    }

    while (bytes_to_write > 0)
    {
        if (bytes_to_write >= (int32_t)sizeof(buf)) {
            res = hb_stor_write_file(phandle, buf, sizeof(buf), &bytes_written);
        } else {
            res = hb_stor_write_file(phandle, buf, bytes_to_write, &bytes_written);
        }
        bytes_to_write -= bytes_written;

        if (res != 0) {
            return HB_FAIL;
        }
    }

    return HB_SUCCESS;
}
/* =============================================================== */


/* returns the number of successfully read bytes */
uint32_t hb_stor_read_buf(f_handle_t* phandle, void* pbuf, uint32_t bytes)
{
    uint32_t bytes_read;
    uint32_t res;

    res = hb_stor_read_file(phandle, pbuf, bytes, &bytes_read);

    if (res != HB_SUCCESS) {
        hb_stor_close_file(phandle);
        DBG_STOR2(HBPRINT("### ERR: hb_stor_read_buf() FAILED [%ld], %ld bytes read\n\r", _error_code, bytes_read);)
        return 0;
    }

    return bytes_read;
}
/* =============================================================== */


/* returns the number of successfully read bytes */
uint32_t hb_stor_read_buf_from_offset(f_handle_t* phandle, void* pbuf, uint32_t bytes, uint32_t offset)
{
    uint32_t wr_offset = hb_stor_tell_file(phandle);
    uint32_t bytes_read;
    uint32_t res;

    hb_stor_seek_file(phandle, offset);

    res = hb_stor_read_file(phandle, pbuf, bytes, &bytes_read);

    if (res != HB_SUCCESS) {
        hb_stor_close_file(phandle);
        DBG_STOR2(HBPRINT("### ERR: hb_stor_read_buf() FAILED [%ld], %ld bytes read\n\r", res, bytes_read);)
        return 0;
    }
    hb_stor_seek_file(phandle, wr_offset);

    return bytes_read;
}
/* =============================================================== */


#define WRITE_CHUNK_SIZE 256


/* returns the number of successfully written bytes */
uint32_t hb_stor_write_buf(f_handle_t* phandle, void* pbuf, uint32_t bytes)
{
    uint32_t total_written = 0;
    uint32_t res;

    uint8_t* pfrom = pbuf;
    uint32_t bytes_to_write;
    uint32_t bytes_written;


/*
        // AK NOTE: for some reason this is a very slow method of writing large buffers.
        //          For example, it takes > 3000 ms to write a 40 KB buffer, while the
        //          implemented below realization does it 10 times faster!!! (300 ms)
        //          In addition, chunk size 256 works better (faster) than 512 (302 vs. 345 ms)

    res = f_write(phandle->handle, pbuf, bytes, (UINT*)&bytes_written);

    if (0 != res) {
        hb_stor_close_file(phandle);
        DBG_ERR(printf("### ERR: hb_stor_write_buf() - f_write() FAILED [%d]\n", res);)
        return 0;
    }

    total_bytes = bytes_written;
*/

    /* fool-tolerance */
    if (bytes == 0) {
        _error_code = 0;
        return 0;
    }

    /* optimized write-loop */
    do
    {
        if ((total_written + WRITE_CHUNK_SIZE) < bytes) {
            bytes_to_write = WRITE_CHUNK_SIZE;
        } else {
            bytes_to_write = bytes - total_written;
        }

        res = hb_stor_write_file(phandle, &pfrom[total_written], bytes_to_write, &bytes_written);

        if (res != HB_SUCCESS) {
            hb_stor_close_file(phandle);
            DBG_ERR(HBPRINT("### ERR: hb_stor_write_buf() - f_write() FAILED [%ld]\n", _error_code);)
            return 0;
        }

        if (bytes_written != bytes_to_write)
        {
            DBG_WARN(HBPRINT("** WARN: hb_stor_write_buf() - error writing log file [%ld vs %ld]\n",
                                bytes_written, bytes_to_write);)
            break;
        }

        total_written += bytes_written;

    } while (total_written < bytes);

    res = hb_stor_sync_file(phandle);  // AK TODO: do we always need to do sync?
    if (HB_SUCCESS != res) {
        DBG_ERR(HBPRINT("### ERR: hb_stor_write_buf() - f_sync() FAILED [%ld]\n", _error_code);)
        //return 0;
        // TODO: to come up with the correct behavior
    }

    return total_written;
}
/* =============================================================== */


/* returns HB_SUCCESS or HB_FAIL */
uint32_t hb_stor_set_rw_pos(f_handle_t* phandle, uint32_t offset)
{
    return hb_stor_seek_file(phandle, offset);
}
/* =============================================================== */


int32_t hb_stor_get_file_size(f_handle_t* phandle)
{
    if (phandle->handle != NULL)    // for ability to safely call this function more than once
    {
        return hb_stor_size_file(phandle);
    }

    return HB_ERROR;
}

/* =============================================================== */

// AK NOTE: code formatting (indents) adjusted to be in accordance with the original file

// AK TODO: consider a more meaningful name for this function (now we read not only log files):
//       hb_stor_read_file_fragment()?
//       hb_stor_read_file_data()?
uint32_t hb_stor_read_log_portion(f_handle_t *phandle, const char *fname, uint32_t start_address, uint32_t bytes, void *buffer, uint32_t *bytes_read)
{
    uint8_t bOpen = HB_FALSE;
    f_handle_t handle = {NULL};

    if(!buffer || !bytes_read)
        return HB_ERROR;

    if(!phandle && !fname)
        return HB_ERROR;

    if(!phandle)
        phandle = &handle;

    if(phandle->handle == NULL)
    {
        if (hb_stor_open_file(phandle, fname) == HB_SUCCESS)
            bOpen = HB_TRUE;
        else
            return HB_ERROR;
    }

    if( (*bytes_read = hb_stor_read_buf_from_offset(phandle, buffer, bytes, start_address)) != bytes)
    {
        if(bOpen)
            hb_stor_close_file(phandle);
        return HB_FAIL;
    }

    if(bOpen)
        hb_stor_close_file(phandle);
    return HB_SUCCESS;
}
/* =============================================================== */


    // NOTE: hb_stor_get_records_count() is designed to work for both: already opened
    //       and not yet opened files. It's primarily designed to be used in the context
    //       of log-file reporting routines, when we don't know if the file of our interest
    //       is open or not.
    //       For files that for sure are known as already opened we can read the number
    //       of available records from the 'recs_count' field of their 'f_handle_t' object

uint32_t hb_stor_get_records_count(f_handle_t *pDat_handle, f_handle_t *pCfg_handle, const char *dat_fname, const char *cfg_fname)
{
    uint32_t recs_count = 0;
    f_handle_t dhandle = {NULL};
    f_handle_t chandle = {NULL};

    if ((pDat_handle == NULL || pCfg_handle == NULL) && (dat_fname == NULL || cfg_fname == NULL)) {
        // AK TODO: should we return an error code here? Seems to be a good idea...
        return 0;
    }

    if (pDat_handle == NULL) {
        pDat_handle = &dhandle;
    }

    if (pCfg_handle == NULL) {
        pCfg_handle = &chandle;
    }

    if (pDat_handle->handle != NULL)
    {
        recs_count = pDat_handle->recs_count;
    }
    else
    {
        // AK TODO: count of records is stored in the .cfg file, so we DO NOT even need to open
        //          the .dat file here. We can just return the number of records read from the
        //          .cfg file

        //if ((hb_stor_open_file(pDat_handle, dat_fname) == HB_SUCCESS) &&
        //    (hb_stor_open_file(pCfg_handle, cfg_fname) == HB_SUCCESS))

        if (hb_stor_open_file(pCfg_handle, cfg_fname) == HB_SUCCESS)
        {
            uint32_t bytes_read = hb_stor_read_buf(pCfg_handle, &recs_count, sizeof(uint32_t));

            if (bytes_read != sizeof(uint32_t)) {
                // AK TODO: again, should we return an error code here?
                recs_count = 0;
            }

            hb_stor_close_file(pDat_handle);
            hb_stor_close_file(pCfg_handle);
        }
    }

    return recs_count;
}
/* =============================================================== */

int32_t hb_stor_get_error_code(void)
{
    return _error_code;
}
/* =============================================================== */

/* =============================================================== */

uint32_t hb_stor_get_file_info(const char *path, file_info_t *fno)
{
    return hb_fs.stat((void*)&hb_fs_context, path, fno);
}
/* =============================================================== */

uint32_t hb_stor_rename(const char *path_old, const char *path_new)
{
    return hb_fs.rename((void*)&hb_fs_context, path_old, path_new);
}
/* =============================================================== */

uint32_t hb_stor_trunc_file(f_handle_t* phandle, const uint32_t size)
{
    return hb_fs.trunc((void*)&hb_fs_context, phandle->handle, size);
}
/* =============================================================== */
#endif // HB_ATMEL_ARM





/*************************************************************************
*  Implementation of file system
*************************************************************************/

#if defined(HB_ATMEL_ARM) || defined(EMULATE_TARGET_FS)

#if defined (FAT_FS)

uint32_t _fs_init_cfg(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;

#ifdef W25QXX_MEMORY
    fsc->memory_lun_id = LUN_ID_W25QXX_MEM;
#else
    fsc->memory_lun_id = LUN_ID_SD_MMC_0_MEM;
#endif
}

uint32_t _fs_mount(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;
    DBG_STOR(HBPRINT("File system is FAT_FS\r\n");)
    hal_memset(fsc->pfs, 0, sizeof(FATFS));

    FRESULT res = f_mount(fsc->memory_lun_id, fsc->pfs);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_mount err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_unmount(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_mount(fsc->memory_lun_id, NULL);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_unmount err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_open(void *pfs_context, HB_FIL *pfile, const char* pfname, uint32_t mode)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_open(pfile, pfname, mode);
    if (res == FR_NO_FILE) {
        return HB_STOR_ERR_NO_FILE;
    }
    else if (res != FR_OK) {
        _error_code = res;
        DBG_WARN(HBPRINT("** WARN: _fs_open err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_format(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_mkfs(0, 0, 0);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_format err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_seek(void *pfs_context, HB_FIL *pfile, const uint32_t offset)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_lseek(pfile, offset);
    if (res != FR_OK) {
        _error_code = res;
        DBG_WARN(HBPRINT("** WARN: _fs_seek err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_write(void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbw)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_write(pfile, pbuf, size, pbw);
    if (res != FR_OK) {
        _error_code = res;
        DBG_WARN(HBPRINT("** WARN: _fs_write err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_read(void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbr)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_read(pfile, pbuf, size, pbr);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_read err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_clear(void *pfs_context, HB_FIL *pfile)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_truncate(pfile);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_clear err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_close(void *pfs_context, HB_FIL *pfile)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_close(pfile);
    if (res != FR_OK) {
        _error_code = res;
        DBG_WARN(HBPRINT("** WARN: _fs_close err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_delete(void *pfs_context, const char* pfname)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_unlink(pfname);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_delete err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_tell(void *pfs_context, HB_FIL *pfile, uint32_t *ps)
{
    *ps = f_tell(pfile);
    return HB_SUCCESS;
}

uint32_t _fs_sync(void *pfs_context, HB_FIL *pfile)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_sync(pfile);
    if (res != FR_OK) {
        _error_code = res;
        DBG_WARN(HBPRINT("** WARN: _fs_sync err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_get_size(void *pfs_context, HB_FIL *pfile, uint32_t *psize)
{
    hb_fs_context_t *fsc = pfs_context;
    *psize = f_size(pfile);
    return HB_SUCCESS;
}

uint32_t _fs_trunc_file(void *pfs_context,  HB_FIL *pfile, const uint32_t size)
{
    hb_fs_context_t *fsc = pfs_context;
    FRESULT res = f_lseek(pfile, size);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_seek err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

uint32_t _fs_stat(void *pfs_context, const char *ppath, file_info_t *pfno)
{
    FILINFO fno;
    FRESULT res = f_stat(ppath, &fno);
    if (res == FR_NO_FILE) {
        return HB_STOR_ERR_NO_FILE;
    }
    else if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_stat err...[%d]\r\n", res);)
        return HB_FAIL;
    }

    pfno->size = fno.fsize;
    // TODO: fill other fields

    return HB_SUCCESS;
}

uint32_t _fs_rename(void *pfs_context, const char *ppath_old, const char *ppath_new)
{
    FRESULT res = f_rename(ppath_old, ppath_new);
    if (res != FR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_rename err...[%d]\r\n", res);)
        return HB_FAIL;
    }
    return HB_SUCCESS;
}

#endif // FAT_FS

#if defined (LITTLE_FS)

#if !defined(EMULATE_TARGET_FS)
#include "HW/hw_w25q.h"

uint32_t _fs_init_cfg(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;

    const w25qxx_info_t *p_w25_info = hw_w25qxx_get_info();
    // block device operations
    fsc->pcfg->read  = spi_mem_read;
    fsc->pcfg->prog  = spi_mem_prog;
    fsc->pcfg->erase = spi_mem_erase;
    fsc->pcfg->sync  = spi_mem_sync;

    // block device configuration
    fsc->pcfg->read_size   = p_w25_info->page_size;
    fsc->pcfg->prog_size   = p_w25_info->page_size;
    fsc->pcfg->block_size  = p_w25_info->sector_size;
    fsc->pcfg->block_count = p_w25_info->sector_count;
    fsc->pcfg->block_cycles = 100;
    fsc->pcfg->cache_size  = fsc->pcfg->prog_size;
    fsc->pcfg->lookahead_size = p_w25_info->page_size;

    //DBG_WARN(HBPRINT("SPI Flash info\r\n");)
    //DBG_WARN(HBPRINT("** Read size...[%d]\r\n", fsc->pcfg->read_size );)
    //DBG_WARN(HBPRINT("** Prog size...[%d]\r\n", fsc->pcfg->prog_size);)
    //DBG_WARN(HBPRINT("** Block size...[%d]\r\n", fsc->pcfg->block_size);)
    //DBG_WARN(HBPRINT("** Block count...[%d]\r\n", fsc->pcfg->block_count);)
    //DBG_WARN(HBPRINT("** Lookahead...[%d]\r\n", fsc->pcfg->lookahead_size);)

    return HB_SUCCESS;
}

#else

// fvs: ToDo:
#include "emubd/lfs_emubd.h"

uint32_t _fs_init_cfg(void *pfs_context) {
    hb_fs_context_t *fsc = pfs_context;

	static lfs_emubd_t emu;

    // block device operations
	fsc->pcfg->context = &emu;
    fsc->pcfg->read  = lfs_emubd_read;
    fsc->pcfg->prog  = lfs_emubd_prog;
    fsc->pcfg->erase = lfs_emubd_erase;
    fsc->pcfg->sync  = lfs_emubd_sync;

    // block device configuration
    fsc->pcfg->read_size   = 256;
    fsc->pcfg->prog_size   = 256;
    fsc->pcfg->block_size  = 4096;
    fsc->pcfg->block_count = 4096;
    fsc->pcfg->block_cycles = 100;
    fsc->pcfg->cache_size  = fsc->pcfg->prog_size;
    fsc->pcfg->lookahead_size = 256;

	lfs_emubd_create(fsc->pcfg, "emubd");

    return HB_SUCCESS;
}

#endif


uint32_t _fs_mount(void *pfs_context)
{
    DBG_STOR(HBPRINT("File system is LITTLE_FS\r\n");)

    hb_fs_context_t *fsc = pfs_context;
    int err = 0;
    
    if (fsc->mounted) {
        return HB_SUCCESS;
    }

    err = lfs_mount(fsc->pfs, fsc->pcfg);
    if (LFS_ERR_OK != err) {
        /*
        lfs_format(fsc->pfs, fsc->pcfg);
        err = lfs_mount(fsc->pfs, fsc->pcfg);
        if (LFS_ERR_OK != err)
        */ {
            DBG_WARN(HBPRINT("** WARN: _fs_mount err...[%d]\r\n", err);)
            return HB_FAIL;
        }
    }
    fsc->mounted = HB_TRUE;

    return HB_SUCCESS;
}

uint32_t _fs_unmount(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;
    if (!fsc->mounted) {
        return HB_SUCCESS;
    }        
    int err = lfs_unmount(fsc->pfs);
    fsc->mounted = HB_FALSE;
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    DBG_WARN(HBPRINT("** WARN: _fs_unmount err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_open(void *pfs_context, HB_FIL *pfile, const char* pfname, uint32_t mode)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_file_open(fsc->pfs, pfile, pfname, mode);
    if (err == LFS_ERR_NOENT) {
        return HB_STOR_ERR_NO_FILE;
    }
    else if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    _error_code = err;
    DBG_WARN(HBPRINT("** WARN: _fs_open err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_format(void *pfs_context)
{
    hb_fs_context_t *fsc = pfs_context;


    Bool mounted = fsc->mounted;

    if (mounted) {
        _fs_unmount(pfs_context);
    } else {
        // Just in case format is called on uninitialized fs
        _fs_init_cfg(pfs_context);
    }

    int err = lfs_format(fsc->pfs, fsc->pcfg);

    if (err < LFS_ERR_OK) {
        DBG_WARN(HBPRINT("** WARN: _fs_format err...[%d]\r\n", err);)
        return HB_FAIL;
    }

    if (mounted) {
        return _fs_mount(pfs_context);
    }

    return HB_SUCCESS;
}

uint32_t _fs_seek(void *pfs_context, HB_FIL *pfile, const uint32_t offset)
{
    hb_fs_context_t *fsc = pfs_context;
    //int32_t size = lfs_file_size(fsc->pfs, pfile);
    //if (size >= 0) {
        //if (offset > size) { //resized file like in FAT_FS
            //lfs_file_truncate(fsc->pfs, pfile, offset);
        //}
    //}
    //else {
        //DBG_WARN(HBPRINT("** WARN: _fs_seek err...[%d]\r\n", size);)
        //return HB_FAIL;
    //}

    int32_t err = lfs_file_seek(fsc->pfs, pfile, offset, LFS_SEEK_SET);
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    else {
        _error_code = err;
        DBG_WARN(HBPRINT("** WARN: _fs_seek err...[%ld]\r\n", err);)
        return HB_FAIL;
    }
}

uint32_t _fs_write(void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbw)
{
    hb_fs_context_t *fsc = pfs_context;
    int32_t err = lfs_file_write(fsc->pfs, pfile, pbuf, size);
    if (err >= LFS_ERR_OK) {
        *pbw = err;
        return HB_SUCCESS;
    }
    _error_code = err;
    DBG_WARN(HBPRINT("** WARN: _fs_write err...[%ld]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_read(void *pfs_context, HB_FIL *pfile, uint8_t *pbuf, uint32_t size, uint32_t *pbr)
{
    hb_fs_context_t *fsc = pfs_context;
    int32_t err = lfs_file_read(fsc->pfs, pfile, pbuf, size);
    if (err >= LFS_ERR_OK) {
        *pbr = err;
        return HB_SUCCESS;
    }
    _error_code = err;
    DBG_WARN(HBPRINT("** WARN: _fs_read err...[%ld]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_clear(void *pfs_context, HB_FIL *pfile)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_file_truncate(fsc->pfs, pfile, 0);
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    DBG_WARN(HBPRINT("** WARN: _fs_clear err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_close(void *pfs_context, HB_FIL *pfile)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_file_close(fsc->pfs, pfile);
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    _error_code = err;
    DBG_WARN(HBPRINT("** WARN: _fs_close err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_delete(void *pfs_context, const char* pfname)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_remove(fsc->pfs, pfname);
    if ((err >= LFS_ERR_OK) || (err == LFS_ERR_NOENT)) {
        return HB_SUCCESS;
    }
    DBG_WARN(HBPRINT("** WARN: %s error at file %s [%d]\n", __FUNCTION__, pfname, err);)
    return HB_FAIL;
}

uint32_t _fs_tell(void *pfs_context, HB_FIL *pfile, uint32_t *ps)
{
    hb_fs_context_t *fsc = pfs_context;
    int32_t err = lfs_file_tell(fsc->pfs, pfile);
    if (err >= LFS_ERR_OK) {
        *ps = err;
        return HB_SUCCESS;
    }
    DBG_WARN(HBPRINT("** WARN: _fs_tell err...[%ld]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_sync(void *pfs_context, HB_FIL *pfile)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_file_sync(fsc->pfs, pfile);
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    _error_code = err;
    DBG_WARN(HBPRINT("** WARN: _fs_sync err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_get_size(void *pfs_context, HB_FIL *pfile, uint32_t *psize)
{
    hb_fs_context_t *fsc = pfs_context;
    int32_t err = lfs_file_size(fsc->pfs, pfile);
    if (err >= LFS_ERR_OK) {
        *psize = err;
        return HB_SUCCESS;
    }
    *psize = 0;
    DBG_WARN(HBPRINT("** WARN: _fs_get_size err...[%ld]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_trunc_file(void *pfs_context,  HB_FIL *pfile, const uint32_t size)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_file_truncate(fsc->pfs, pfile, size);
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    DBG_WARN(HBPRINT("** WARN: _fs_trunc_file err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_stat(void *pfs_context, const char *ppath, file_info_t *pfno)
{
    hb_fs_context_t *fsc = pfs_context;
    struct lfs_info fno = {0};

    int err = lfs_stat(fsc->pfs, ppath, &fno);
    if (err == LFS_ERR_NOENT) {
        return HB_STOR_ERR_NO_FILE;
    }
    else if (err >= LFS_ERR_OK) {
        pfno->size = fno.size;
        // TODO: fill other fields
        return HB_SUCCESS;
    }

    DBG_WARN(HBPRINT("** WARN: _fs_stat err...[%d]\r\n", err);)
    return HB_FAIL;
}

uint32_t _fs_rename(void *pfs_context, const char *ppath_old, const char *ppath_new)
{
    hb_fs_context_t *fsc = pfs_context;
    int err = lfs_rename(fsc->pfs, ppath_old, ppath_new);
    if (err >= LFS_ERR_OK) {
        return HB_SUCCESS;
    }
    DBG_WARN(HBPRINT("** WARN: _fs_rename err...[%d]\r\n", err);)
    return HB_FAIL;
}
#endif // LITTLE_FS

#endif // HB_ATMEL_ARM
/* ==================================================================== */
